(* 不要定義削除 *)

open Shared
open Id
open Global
open ILang

let labels = ref S2.empty
let add = function Label l -> labels := S2.add l !labels | _ -> ()

(* 特定の分岐先でしか使われず、移動してもよさそうな定義を分岐先に移動する *)
let rec insert x e1 e2 =
  match e2 with
  | V _ | Op _ -> Let(x,e1,e2)
  | If(c,y,z,e3,e4) ->
      let fv3 = fv e3 and fv4 = fv e4 in
      if S.mem x fv3 then
        if S.mem x fv4 then If(c,y,z,Let(x,e1,e3),Let(x,e1,e4))
        else If(c,y,z,Let(x,e1,e3),e4)
      else If(c,y,z,e3,Let(x,e1,e4))
  | Let(y,e3,e4) ->
      if S.mem x (fv e3) then Let(y, insert x e1 e3, e4)
      else Let(y,e3, insert x e1 e4)
  | LetRec(y,zs,e3,e4) ->
      if S.mem x (fv e3) then LetRec(y,zs, insert x e1 e3, e4)
      else LetRec(y,zs,e3, insert x e1 e4)

(* 引数の env は副作用がないとわかる関数の集合 *)
(* 返り値は (変換後の式、生存変数、どの分岐先でも使われる変数、
   副作用があるかどうか) の組 *)
let rec g env e =
  match e with
  | V x -> add x; let s = fv0 x in (e, s, s, false)
  | Op(Label l, xs) ->
      labels := S2.add l !labels;
      List.iter add xs;
      let s = fv1 xs in
      (e, s, s,
       List.mem l Builtin.side_effective_funs ||
       (not (List.mem l Builtin.fun_names) &&
        not (S2.mem l !SideEffect.result)))
  | Op(Var x,ys) ->
      List.iter add ys;
      let s = S.add x (fv1 ys) in
      (e, s, s, not (S.mem x env))
  | Op _ -> assert false
  | If(c,x,y,e1,e2) ->
      let e1_, liv1, liv1_, se1 = g env e1 in
      let e2_, liv2, liv2_, se2 = g env e2 in
      let s = S.union (fv0 x) (fv0 y) in
      (If(c,x,y,e1_,e2_),
       S.union s (S.union liv1 liv2),
       S.union s (S.inter liv1_ liv2_),
       se1 || se2)
  | Let(x,e1,e2) ->
      let e1_, liv1, liv1_, se1 = g env e1 in
      let e2_, liv2, liv2_, se2 = g env e2 in
      if se1 || S.mem x liv2_ then
        let e_ = match e2_ with
        | V(Var y) when Id.equal x y -> e1_
        | _ -> Let(x,e1_,e2_)
        in
        let liv2 = S.remove x liv2 in
        let s = S.union liv1 liv2 in
        let s_ =
          if se1 then s (* 副作用はキモいので日和る *)
          else S.union (S.inter liv1 liv2) (S.union liv1_ liv2_)
        in
        (e_, s, s_, se1 || se2)
      else if S.mem x liv2 then begin
        Format.eprintf "%s partially unused@." (Id.to_string x);
        (insert x e1_ e2_,
         S.union liv1 (S.remove x liv2),
         liv2_,
         se2)
      end else begin
        Format.eprintf "%s eliminated@." (Id.to_string x);
        (e2_, liv2, liv2_, se2)
      end
  | LetRec(x,ys,e1,e2) ->
      let e1_, liv1, liv1_, se1 = g env e1 in
      let env_ = if se1 then env else S.add x env in
      let e2_, liv2, liv2_, se2 = g env_ e2 in
      if S.mem x liv2 then
        let liv1 = S.remove x (S.remove_list ys liv1) in
        let liv2 = S.remove x liv2 in
        (LetRec(x,ys,e1_,e2_),
         S.union liv1 liv2,
         S.union (S.inter liv1 liv2) (S.union liv1_ liv2_),
         se2)
      else begin
        Format.eprintf "%s eliminated@." (Id.to_string x);
        (e2_, liv2, liv2_, se2)
      end

let g e = let e_, _,_,_ = g S.empty e in Alpha.f e_

let f e =
  (* ignore(SideEffect.f e) をここで呼ぶ方が安全ではある *)
  labels := S2.empty;
  let e = g e in
  let rec iter src dest =
    match List.partition (fun { name = x } -> S2.mem x !labels) src with
    | [],_ -> dest
    | used,rest -> iter rest
          (List.fold_left
             (fun acc x -> { x with body = g x.body } :: acc)
             dest used)
  in
  gl_funs := iter !gl_funs [];
  Hashtbl.iter (fun k vs -> Array.iter add vs) gl_vars;
  Hashtbl.iter
    (fun k _ -> if not (S2.mem k !labels) then Hashtbl.remove gl_vars k)
    gl_vars;
  labels := S2.empty;
  e
