%{
open Syntax
%}

/* トークンの宣言 */
%token <bool> BOOL
%token <int> INT
%token <float> FLOAT
%token <string> IDENT
%token BARBAR
%token AMPAMP
%token <string> OP1 /* & | */
%token <string> OP2 /* = < > */
%token <string> OP3 /* + - 単項で使用可 */
%token <string> OP4 /* * / % */
%token IF
%token THEN
%token ELSE
%token LET
%token EQUAL
%token IN
%token REC
%token COMMA
%token DOT
%token LESS_MINUS
%token MINUS_GREATER
%token EXTERN
%token COLON
%token SEMICOLON
%token SEMISEMI
%token LPAREN
%token RPAREN
%token EOF

/* 優先順位と結合 */
%right _let_
%right SEMICOLON
%right _if_
%right LESS_MINUS MINUS_GREATER
%left COMMA
%left BARBAR
%left AMPAMP
%left OP1
%left OP2 EQUAL
%left OP3
%left OP4
%right _unary_op3_

/* 開始記号の宣言 */
%type <Syntax.t> exp
%start exp

%type <Syntax.topdef list> main
%start main

%%

simple_exp: /* 括弧をつけなくても関数の引数になれる式 */
| LPAREN exp RPAREN
    { $2 }
| LPAREN RPAREN
    { Unit }
| BOOL
    { Bool($1) }
| INT
    { Int($1) }
| FLOAT
    { Float($1) }
| IDENT
    { Var($1) }
| simple_exp DOT LPAREN exp RPAREN
    { Op(Var "->", [$1; $4]) }

exp: /* 一般の式 */
| simple_exp
    { $1 }
| OP3 exp
    %prec _unary_op3_
    { match $1,$2 with
      | "-", Float f -> Float(-.f) (* -1.0などは型エラーではないことにする *)
      | "-", e -> Op(Var "-", [Int 0; e])
      | "-.", e -> Op(Var "-.", [Float 0.; e])
      | "+", e -> e
      | f, e -> Op(Var f, [e]) }
| exp BARBAR exp
    { If($1, Bool true, $3) }
| exp AMPAMP exp
    { If($1, $3, Bool false) }
| exp OP1 exp
    { Op(Var $2, [$1; $3]) }
| exp EQUAL exp
    { Op(Var "=", [$1; $3]) }
| exp OP2 exp
    { Op(Var $2, [$1; $3]) }
| exp OP3 exp
    { Op(Var $2, [$1; $3]) }
| exp OP4 exp
    { Op(Var $2, [$1; $3]) }
| IF exp THEN exp ELSE exp
    %prec _if_
    { If($2, $4, $6) }
| vdef IN exp
    %prec _let_
    { Let($1, $3) }
| fdef IN exp
    %prec _let_
    { LetRec($1, $3) }
| exp actual_args
    { Op($1, $2) }
| exp_commas
    { Tuple(List.rev $1) }
| simple_exp DOT LPAREN exp RPAREN LESS_MINUS exp
    { Op(Var "<-", [$1; $4; $7]) }
| exp SEMICOLON exp
    { Let(([], $1), $3) }
| exp SEMICOLON /* ひぎぃ */
    { $1 }
| error
    { failwith
        (Printf.sprintf "parse error near characters %d-%d, line %d"
           (Parsing.symbol_start ())
           (Parsing.symbol_end ())
           !Err.linenum) }

vdef:
| LET pat EQUAL exp
    { ($2, $4) }

pat:
| id_commas
    { List.rev $1 }
| LPAREN id_commas RPAREN
    { List.rev $2 }

fdef:
| LET REC IDENT formal_args EQUAL exp
    { ($3, $4, $6) }

formal_args:
| IDENT formal_args
    { $1 :: $2 }
| IDENT
    { [$1] }

actual_args:
| simple_exp actual_args
    { $1 :: $2 }
| simple_exp
    { [$1] }

id_commas:
| IDENT
    { [$1] }
| id_commas COMMA IDENT
    { $3 :: $1 }

exp_commas:
| exp_commas COMMA exp
    { $3 :: $1 }
| exp COMMA exp
    { [$3; $1] }

main:
| topdef
    { [$1] }
| topdef main
    { $1 :: $2 }

topdef:
| topdef2
    { $1 }
| topdef2 SEMISEMI
    { $1 }

topdef2:
| vdef
    { Defvar $1 }
| fdef
    { Defun $1 }
| exp
    { Defvar((["_"], $1)) }
| EXTERN IDENT COLON typedef
    { match $2,$4 with
      | "main",_ -> failwith "function \"main\" cannot be external"
      | x,Type.Tfun _ -> Extern((x, $4))
      | _ -> failwith "external identifiers must be functions" }

typedef:
| IDENT
    { match $1 with
      | "int" -> Type.Tint
      | "float" -> Type.Tfloat
      | "bool" -> Type.Tbool
      | "unit" -> Type.Tunit
      | x -> failwith ("no such type: " ^ x) }
| type_commas
    { Type.Ttuple(List.rev $1) }
| typedef IDENT
    { match $2 with
      | "array" -> Type.Tarray $1
      | x -> failwith ("no such type: " ^ x) }
| typedef MINUS_GREATER typedef
    { match $1 with
      | Type.Ttuple ts -> Type.Tfun(ts, $3)
      | x -> Type.Tfun([x], $3) }
| LPAREN typedef RPAREN
    { $2 }

type_commas:
| typedef COMMA typedef
    { [$3; $1] }
| type_commas COMMA typedef
    { $3 :: $1 }
