open Shared
open List
open TreeAsm

let rec g0 acc e =
  match e with
  | Let(_, If _, _) | Let(_, IfI _, _) | Let(_, FIf _, _) | Let(_, Call _, _)
  | Ret(Mov _) | Ret Nop | Ret(Call _) -> acc, e
  | Ret exp ->
      let x = match exp with St _ | Op1(Wr,_) -> reg_dummy | _ -> new_str () in
      g0 acc (Let(x, exp, Ret(Mov x)))
  | Let(x, exp, e1) -> g0 (e::acc) e1

let nnop = function
  | Op2(FAdd,_,_) | Op2(FSub,_,_) | Op2(FMul,_,_) -> 2
  | Op1i(Ld,_,_) -> 1
  | Op1(FInv,_) | Op1(FSqrt,_) -> 3
  | _ -> 0

let use = function
  | Ret _ -> assert false
  | Let(_,exp,_) -> match exp with
    | Nop | If _ | IfI _ | FIf _ | Call _ | Save _ | Restore _ -> assert false
    | St(x,y,_) | Op2(_,x,y) -> [x;y]
    | Op1i(_,x,_) | Op1(_,x) | Mov x -> [x]
    | Imm _ | Rd -> []
let def = function
  | Ret _ -> assert false
  | Let(_,St(_,_,L r),_) when is_reg r -> [r]
  | Let(x,Op2(Array,_,_),_) -> [x;reg_hp]
  | Let(x,_,_) -> if x = reg_dummy then [] else [x]

let dep0 lst e =
  let u = use e and d = def e in
  fold_left
    (fun acc e2 -> match e2 with Ret _ -> assert false
     | Let(_,exp,_) -> let u2 = use e2 and d2 = def e2 in
       if exists (fun x -> mem x d) u2 ||
         exists (fun x -> mem x d2) u ||
         exists (fun x -> mem x d2) d
       then (e2,nnop exp) :: acc else acc)
    [] lst

let g1 lst e =
  match e with
  | Let(x, exp, _) -> let dep =
      let aux x = x <> reg_zero && is_reg x in
      match exp with
      | If _ | IfI _ | FIf _ | Call _ | Save _ | Restore _ -> assert false
      | Rd -> dep0 lst e @
          map (fun x -> (x,0))
            (filter (function Let(x2,Rd,_) -> x <> x2 | _ -> false) lst)
      | Op1(Wr,y) -> dep0 lst e @
          map (fun e -> e,0)
            (filter (function (Let(_,Op1(Wr,_),_)) -> true | _ -> false) lst)
      | St(y,z,i) -> dep0 lst e @
          map (fun e -> e,0)
            (filter (function Let(_,St(_,z2,i2),_) | Let(_,Op1i(Ld,z2,i2),_) ->
                       aux z || aux z2 || z <> z2 || i = i2
                     | _ -> false)
               lst)
      | Op1i(Ld,y,i) -> dep0 lst e @
          map (fun e -> e,0)
            (filter (function Let(_,St(_,y2,i2),_) ->
                       aux y || aux y2 || y <> y2 || i = i2
                     | _ -> false)
               lst)
      | Op2 _ | Op1i _ | Op1 _ | Mov _ | Imm _ | Nop -> dep0 lst e
    in
    (e, dep)
  | _ -> assert false

let rec g2 acc lst =
  match lst with
  | [] -> acc
  | hd::tl -> g2 (g1 tl hd :: acc) tl

let dep_ok prev (e,d) =
  let rec ntl n l = if n <= 0 then l else match l with
    | [] -> []
    | hd::tl -> ntl (n-1) tl
  in
  memq e (ntl d prev)

let g3 rest prev =
  filter (fun (e,dep) -> for_all (dep_ok prev) dep) rest

let rec f0 e =
  match g0 [] e with
  | [], Ret _ -> e
  | [], Let(x, If(c,y,z,e1,e2), e3) ->
      Let(x, If(c,y,z, f0 e1, f0 e2), f0 e3)
  | [], Let(x, IfI(c,y,i,e1,e2), e3) ->
      Let(x, IfI(c,y,i, f0 e1, f0 e2), f0 e3)
  | [], Let(x, FIf(c,y,z,e1,e2), e3) ->
      Let(x, FIf(c,y,z, f0 e1, f0 e2), f0 e3)
  | [], Let(x, Call(y,zs), e3) ->
      Let(x, Call(y,zs), f0 e3)
  | [], _ -> assert false
  | lst, e -> g (g2 [] lst) [] (f0 e)
and g graph prev cont =
  match graph with
  | [] -> let rec fin p c =
      match p with
      | [] -> c
      | Let(x,exp,_)::tl -> fin tl (Let(x,exp,c))
      | _ -> assert false
    in
    fin prev cont
  | _ ->  match g3 graph prev with
    | [] -> g graph (Let(reg_dummy, Nop, Ret Nop)::prev) cont
    | avail ->
        let hd,_ = nth avail (Random.int (length avail)) in
        g (filter (fun (e,_) -> e != hd) graph) (hd::prev) cont

let f funs =
  Random.self_init ();
  map (fun func -> { func with body = f0 func.body }) funs
