(* スタック割り当て *)

open TreeAsm

let stk_max = ref 0

let rec ra_save_needed tail e =
  let rec aux tail exp =
    match exp with
    | Call _ -> not tail
    | If(_,_,_,e1,e2) | IfI(_,_,_,e1,e2) | FIf(_,_,_,e1,e2) ->
        (ra_save_needed tail e1 || ra_save_needed tail e2)
    | _ -> false
  in
  match e with
  | Ret exp -> aux tail exp
  | Let(x,exp,Ret(Mov y)) when x = y -> aux tail exp
  | Let(_,exp,e) -> aux false exp || ra_save_needed tail e

let ra_save_needed e = ra_save_needed true e

let rec g env n e =
  match e with
  | Ret exp -> g_ env n exp
  | Let(x, exp, e) ->
      let env_, n_, e_ = g_ env n exp in
      let env__, n__, e__ = g env_ n_ e in
      (env__, n__, concat e_ x e__)
and g_ env n exp =
  let g_if e1 e2 k =
    let _, _, e1_ = g env n e1 in
    let _, _, e2_ = g env n e2 in
    (env, n, Ret(k e1_ e2_))
  in
  match exp with
  | Save(r, x) ->
      if not (M2.mem x env) then begin
        stk_max := max !stk_max (n+1);
        (M2.add x n env, n+1, Ret(St(r, reg_sp, I n)))
      end else
        (env, n, Ret Nop)
  | Restore x ->
      if not (M2.mem x env) then failwith ("stkAlloc.g_: " ^ x);
      (env, n, Ret(Op1i(Ld, reg_sp, I (M2.find x env))))
  | If(c,x,y,e1,e2) -> g_if e1 e2
        (fun e1_ e2_ -> If(c,x,y,e1_,e2_))
  | IfI(c,x,i,e1,e2) -> g_if e1 e2
        (fun e1_ e2_ -> IfI(c,x,i,e1_,e2_))
  | FIf(c,x,y,e1,e2) -> g_if e1 e2
      (fun e1_ e2_ -> FIf(c,x,y,e1_,e2_))
  | _ -> (env, n, Ret exp)

let g e =
  stk_max := 0;
  let _, _, e_ = g M2.empty 1 e in
  let s = if ra_save_needed e then max 1 !stk_max else !stk_max in
  (e_, s)

let f xs =
  Format.eprintf "stack allocation...@.";
  List.map (fun { name = x; args = rs; body = e } ->
  let e_, s = g e in
  { name = x; args = rs; body = e_; stksiz = s }) xs
