(* FlatSyntax.t -> ILang.t *)

open Type
open Id
open FlatSyntax
open ILang

let h v =
  match v with
  | FlatSyntax.Unit -> Unit
  | FlatSyntax.Int i -> Int i
  | FlatSyntax.Float f -> Float f
  | Bool b -> Int(if b then 1 else 0)
  | FlatSyntax.Var x -> Var x
let h1 = function FlatSyntax.Eq -> Eq | FlatSyntax.LE -> LE

let rec g e =
  match e with
  | FlatSyntax.V v -> V(h v)
  | FlatSyntax.Tuple vs -> Op(Label ".tuple", List.map h vs)
  | FlatSyntax.Op(x, ys) -> Op(h x, List.map h ys)
  | FlatSyntax.If(c, x, y, e1, e2) -> If(h1 c, h x, h y, g e1, g e2)
  | FlatSyntax.Let(_::_::_ as xs, e1, e2) ->
      let t = Type.Ttuple(List.map (fun x -> x.typ) xs) in
      let tuple = new_id_t "tuple" t in
      let rec iter i xs =
        match xs with
        | [] -> g e2
        | x::xs -> Let(x, Op(Label ".t->", [Var tuple; Int i]), iter (i+1) xs)
      in
      Let(tuple, g e1, iter 0 xs)
  | FlatSyntax.Let([x], e1, e2) -> Let(x, g e1, g e2)
  | FlatSyntax.Let([], e1, e2) -> Let(new_id_t "unit" Tunit, g e1, g e2)
  | FlatSyntax.LetRec(x, ys, e1, e2) -> LetRec(x, ys, g e1, g e2)
  | FlatSyntax.ExtApp(x, ys) -> Op(Label x.name, List.map h ys)

let f e = g e
